using System;
using System.Collections;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Globalization;
using System.Text;
using System.Windows.Forms;
using C1.Win.C1FlexGrid;

namespace C1FlexCalendar
{
    public interface ITaskItem
    {
        string Task { get; }
        DateTime Start { get; }
        DateTime Finish { get; }
    }

    public enum ViewEnum
    {
        Monthly = 0,
        Weekly = 1
    }

    [LicenseProvider(typeof(System.ComponentModel.LicenseProvider))]
    public partial class C1FlexCalendar : C1FlexGrid
    {
        public event EventHandler DateChanged;

        #region fields
        protected Label _btnPrev = new Label();
        protected Label _btnNext = new Label();
        protected Day _first = Day.Default;
        protected bool _showTitle = true;
        protected bool _showDayNames = true;
        protected SelectionRange _selRange = new SelectionRange(DateTime.Today, DateTime.Today);
        protected SelectionRange _dispRange = new SelectionRange(DateTime.Today, DateTime.Today);
        protected DateTime _calendarStart = DateTime.MinValue;
        protected DateTime _calendarMonth = DateTime.Today;
        protected DateTime _timeStart = DateTime.MinValue;
        protected DateTime _timeEnd = DateTime.MinValue;
        protected ViewEnum _view = ViewEnum.Monthly;
        protected ArrayList _tasks = null;
        protected Brush _whiteBrush = new SolidBrush(Color.White);
        protected Brush _khakiBrush = new SolidBrush(Color.Khaki);
        protected DailyTaskItemComparer _dailyComp = new DailyTaskItemComparer();
        protected HourlyTaskItemComparer _hourlyComp = new HourlyTaskItemComparer();
        #endregion

        #region ctor
        public C1FlexCalendar()
        {
            // disable resizing and sorting
            AllowResizing = AllowResizingEnum.None;
            AllowSorting = AllowSortingEnum.None;

            // enable owner draw
            DrawMode = DrawModeEnum.OwnerDraw;

            // add navigation buttons
            _btnPrev.ImageAlign = _btnNext.ImageAlign = ContentAlignment.MiddleCenter;
            _btnPrev.BackColor = _btnNext.BackColor = Color.Transparent;
            _btnPrev.Click += new EventHandler(_btnPrevNext_Click);
            _btnNext.Click += new EventHandler(_btnPrevNext_Click);
            _btnPrev.DoubleClick += new EventHandler(_btnPrevNext_Click);
            _btnNext.DoubleClick += new EventHandler(_btnPrevNext_Click);
            _btnPrev.Font = new Font("Wingdings 3", 16);
            _btnNext.Font = _btnPrev.Font;
            _btnPrev.Text = "|"; // left triangle
            _btnNext.Text = "}"; // right triangle
            Controls.Add(_btnPrev);
            Controls.Add(_btnNext);

            // create styles for owner-drawn cells
            SetupStyles();

            // finish initialization
            InitializeComponent();
        }
        #endregion

        #region object model
        [
        DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)
        ]
        public SelectionRange SelectionRange
        {
            get { return _selRange; }
            set
            {
                _selRange = value;
                BuildCalendar();
            }
        }

        [
        Browsable(false),
        DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)
        ]
        public DateTime SelectionStart
        {
            get { return SelectionRange.Start; }
            set
            {
                SelectionRange.Start = value;
                BuildCalendar();
            }
        }

        [
        Browsable(false),
        DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)
        ]
        public DateTime SelectionEnd
        {
            get { return SelectionRange.End; }
            set
            {
                SelectionRange.End = value;
                BuildCalendar();
            }
        }

        [
        Browsable(false),
        DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)
        ]
        public DateTime TimeStart
        {
            get { return _timeStart; }
        }

        [
        Browsable(false),
        DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)
        ]
        public DateTime TimeEnd
        {
            get { return _timeEnd; }
        }

        [DefaultValue(Day.Default)]
        public Day FirstDayOfWeek
        {
            get { return _first; }
            set
            {
                _first = value;
                BuildCalendar(true);
            }
        }

        [DefaultValue(true)]
        public bool ShowTitle
        {
            get { return _showTitle; }
            set
            {
                _showTitle = value;
                UpdateLayout();
                BuildCalendar();
            }
        }

        [DefaultValue(true)]
        public bool ShowDayNames
        {
            get { return _showDayNames; }
            set
            {
                _showDayNames = value;
                UpdateLayout();
                BuildCalendar();
            }
        }

        [DefaultValue(ViewEnum.Monthly)]
        public ViewEnum View
        {
            get { return _view; }
            set
            {
                _view = value;
                UpdateLayout(true);
                BuildCalendar(true);
            }
        }

        [
		Browsable(false),
        DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)
        ]
        public ArrayList TaskList
        {
            get { return _tasks; }
            set
            {
                _tasks = value;
                _tasks.Sort(_hourlyComp);
                ClearWeeklyView();
                BuildCalendar(true);
            }
        }

        [
		Browsable(false),
        DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)
        ]
        public ArrayList SelectedTaskList
        {
            get
            {
                ArrayList a = new ArrayList();

                for (int c = Col; c <= ColSel; c++)
                {
                    for (int r = Row; r <= RowSel; r++)
                    {
                        object data = GetData(r, c);

                        if (data is ITaskItem)
                        {
                            ITaskItem ti = (ITaskItem)data;

                            if (ti != null && !a.Contains(ti))
                                a.Add(ti);
                        }
                    }
                }

                return a;
            }
        }

        public Size GetSelectionSize()
        {
            CellRange sel = this.Selection;
            Rectangle r1 = GetCellRect(sel.TopRow, sel.LeftCol);
            Rectangle r2 = GetCellRect(sel.BottomRow, sel.RightCol);
            return new Size(r2.Right - r1.Left - 1, r2.Bottom - r1.Top - 1);
        }
        #endregion

        #region overrides
        override protected void OnSizeChanged(EventArgs e)
        {
            base.OnSizeChanged(e);
            UpdateLayout();
        }

        override protected void InitLayout()
        {
            base.InitLayout();

            // adjust row/column sizes based on control size
            UpdateLayout(true);
            
            // set cell values and redraw the calendar
            BuildCalendar(true);
        }

        override protected void OnBeforeSelChange(RangeEventArgs e)
        {
            base.OnBeforeSelChange(e);

            if (e.NewRange.IsSingleCell)
            {
                CellRange nr = e.NewRange;
                CellRange mr = GetMergedRange(nr.r1, nr.c1);

                if (mr.r1 != nr.r1 || mr.c1 != nr.c1 || mr.r2 != nr.r2 || mr.c2 != nr.c2)
                {
                    e.Cancel = true;
                    Select(mr);
                }
            }
        }

        override protected void OnAfterSelChange(RangeEventArgs e)
        {
            base.OnAfterSelChange(e);
            int startCol = (_view == ViewEnum.Monthly) ? 0 : 1;

            if (Row > 1 && Col >= startCol && _calendarStart != DateTime.MinValue)
            {
                DateTime t1 = GetCellDate(Row, Col);
                DateTime t2 = GetCellDate(RowSel, ColSel);

                if (DateTime.Compare(t1, t2) <= 0)
                {
                    t2 = t2.AddHours(1);
                    SelectionRange = new SelectionRange(t1, t2);
                    _timeStart = t1;
                    _timeEnd = t2;
                }
                else
                {
                    t1 = t1.AddHours(1);
                    SelectionRange = new SelectionRange(t2, t1);
                    _timeStart = t2;
                    _timeEnd = t1;
                }
            }
        }

        override protected void OnOwnerDrawCell(OwnerDrawCellEventArgs e)
        {
            base.OnOwnerDrawCell(e);
            e.DrawCell();

            if (_view == ViewEnum.Monthly)
            {
                DrawItemDay(e);
            }
            else if (_view == ViewEnum.Weekly)
            {
                DrawItemHour(e);
            }
        }
        #endregion

        #region utilities
        protected DayOfWeek GetFirstDayOfWeek()
        {
            switch (_first)
            {
                case Day.Monday:    return DayOfWeek.Monday;
                case Day.Tuesday:   return DayOfWeek.Tuesday;
                case Day.Wednesday: return DayOfWeek.Wednesday;
                case Day.Thursday:  return DayOfWeek.Thursday;
                case Day.Friday:    return DayOfWeek.Friday;
                case Day.Saturday:  return DayOfWeek.Saturday;
                case Day.Sunday:    return DayOfWeek.Sunday;
            }
            return CultureInfo.CurrentUICulture.DateTimeFormat.FirstDayOfWeek;
        }

        private void SetWeeklyData(bool clear)
        {
            if (clear)
            {
                ClearWeeklyView();

                if (DateChanged != null)
                    DateChanged(this, new System.EventArgs());
            }

            if (_tasks == null)
                return;

            foreach (ITaskItem ti in _tasks)
            {
                int row = ti.Start.Hour + 2;
                int row2 = row;
                TimeSpan ts = ti.Start.Subtract(_calendarStart);
                int col = ts.Days + 1;

                if (col > 0 && col < Cols.Count)
                {
                    ts = ti.Finish - ti.Start;
                    if (ts.Hours > 0) row2 = row + ts.Hours - 1;
                    if (ts.Minutes > 0) row2++;

                    for (int i = row; i <= row2; i++)
                        SetData(i, col, ti);
                }
            }
        }

        private DateTime GetCellDate(int row, int col)
        {
            DateTime date = _calendarStart;

            if (_view == ViewEnum.Monthly)
            {
                date = date.AddDays(7 * (row - 2));
                date = date.AddDays(col);
            }
            else
            {
                date = date.AddDays(col - 1);
                date = date.AddHours(row - 2);
            }

            return date;
        }

        private bool GetOwnerDrawTaskRange(DateTime date, ref int lb, ref int ub)
        {
            lb = ub = -1;

            if (_tasks == null)
                return false;

            lb = ub = _tasks.BinarySearch(date, _dailyComp);

            if (lb < 0)
                return false;

            for (int n = ub + 1; n < _tasks.Count; n++)
            {
                ITaskItem ti = _tasks[n] as ITaskItem;

                if (date > ti.Start || ti.Start >= date.AddDays(1))
                    break;

                ub = n;
            }

            for (int n = lb - 1; n >= 0; n--)
            {
                ITaskItem ti = _tasks[n] as ITaskItem;

                if (date > ti.Start || ti.Start >= date.AddDays(1))
                    break;

                lb = n;
            }

            return true;
        }

		private bool HasOwnerDrawTasks(DateTime date)
		{
			if (_tasks == null)
				return false;

			return _tasks.BinarySearch(date, _dailyComp) >= 0;

		}

        private void DrawItemDay(OwnerDrawCellEventArgs e)
        {
            if (e.Row >= 2)
            {
                DateTime date = GetCellDate(e.Row, e.Col);

                if (date == DateTime.Today)
                {
                    e.Style = Styles["Today"];
                }
                else if (date.Month != _calendarMonth.Month)
                {
                    e.Style = Styles["GrayDays"];
                }
                else if (date.DayOfWeek == DayOfWeek.Saturday || date.DayOfWeek == DayOfWeek.Sunday)
                {
                    e.Style = Styles["Weekends"];
                }

				if (HasOwnerDrawTasks(date))
				{
                    e.Graphics.FillRectangle(_khakiBrush, e.Bounds);
				}

                e.Text = date.Day.ToString();
                e.DrawCell(DrawCellFlags.Content | DrawCellFlags.Border);
            }
        }

        private void DrawItemHour(OwnerDrawCellEventArgs e)
        {
            if (e.Row >= 2 && e.Col > 0)
            {
                object data = GetData(e.Row, e.Col);

                if (data is ITaskItem)
                {
                    ITaskItem ti = (ITaskItem)data;
                    e.Text = ti.Task;

                    CellRange cr = GetMergedRange(e.Row, e.Col);
                    for (int i = cr.r1; i <= cr.r2; i++)
                    {
                        if (Selection.Contains(i, e.Col))
                        {
                            e.DrawCell();
                            return;
                        }
                    }

                    e.Graphics.FillRectangle(_whiteBrush, e.Bounds);
                    e.DrawCell(DrawCellFlags.Border | DrawCellFlags.Content);
                }
            }
        }

        protected void SetupStyles()
        {
            // change alignment and default format for normal style
            CellStyle cs = Styles.Normal;
            cs.TextAlign = TextAlignEnum.CenterCenter;
            cs.Format = "%d";

            // title/navigation bar (row 0)
            cs = Styles.Add("Title");
            cs.BackColor = SystemColors.ActiveCaption;
            cs.ForeColor = SystemColors.ActiveCaptionText;
            cs.Font = new Font(cs.Font.FontFamily.Name, 14, FontStyle.Bold);
            cs.Border.Style = BorderStyleEnum.None;
            cs.Border.Color = cs.BackColor;
            cs.TextAlign = TextAlignEnum.CenterCenter;

            // weekdays (monthly view)
            cs = Styles.Add("DayNames");
            cs.Font = new Font(cs.Font, FontStyle.Regular);
            cs.Border.Style = BorderStyleEnum.Flat;
            cs.Border.Direction = BorderDirEnum.Horizontal;
            cs.Border.Color = cs.ForeColor;
            cs.Border.Width = 1;
            cs.Format = "ddd";
            cs.TextAlign = TextAlignEnum.CenterCenter;

            // weekdays (weekly view)
            cs = Styles.Add("WeekDayNames");
            cs.Font = new Font(cs.Font, FontStyle.Regular);
            cs.Border.Style = BorderStyleEnum.Flat;
            cs.Border.Direction = BorderDirEnum.Both;
            cs.Border.Color = SystemColors.ControlDark;
            cs.Border.Width = 1;
            cs.Format = "ddd\nM/d";
            cs.TextAlign = TextAlignEnum.CenterCenter;

            // hours (weekly view, column 0)
            cs = Styles.Add("Hours");
            cs.Font = new Font(cs.Font, FontStyle.Regular);
            cs.Border.Style = BorderStyleEnum.Flat;
            cs.Border.Direction = BorderDirEnum.Both;
            cs.Border.Color = SystemColors.ControlDark;
            cs.Border.Width = 1;
            cs.Format = "t";
            cs.TextAlign = TextAlignEnum.RightCenter;

            // gray dates
            cs = Styles.Add("GrayDays");
            cs.ForeColor = SystemColors.GrayText;

            // weekends
            cs = Styles.Add("Weekends");
            cs.ForeColor = SystemColors.WindowText;

            // today
            cs = Styles.Add("Today");
            cs.Font = new Font(cs.Font, FontStyle.Bold);

            // free time (weekly view)
            cs = Styles.Add("FreeTime");
            cs.WordWrap = true;
            cs.BackColor = Color.Khaki;
            cs.Border.Style = BorderStyleEnum.Flat;
            cs.Border.Color = SystemColors.Control;
            cs.Border.Direction = BorderDirEnum.Both;
            cs.Border.Width = 1;

            // work time (weekly view)
            cs = Styles.Add("WorkTime");
            cs.WordWrap = true;
            cs.BackColor = Color.LightGoldenrodYellow;
            cs.Border.Style = BorderStyleEnum.Flat;
            cs.Border.Color = SystemColors.Control;
            cs.Border.Direction = BorderDirEnum.Both;
            cs.Border.Width = 1;
        }

        protected void ClearWeeklyView()
        {
            if (_view != ViewEnum.Weekly)
                return;

            CellRange rg = GetCellRange(2, 1, 25, 7);
            rg.Clear(C1.Win.C1FlexGrid.ClearFlags.Content);

            rg = GetCellRange(2, 1, 9, 7);
            rg.Style = Styles["FreeTime"];
            rg = GetCellRange(10, 1, 18, 7);
            rg.Style = Styles["WorkTime"];
            rg = GetCellRange(19, 1, 25, 7);
            rg.Style = Styles["FreeTime"];
        }

        protected void UpdateLayout()
        {
            UpdateLayout(false);
        }

        protected void UpdateLayout(bool viewChanged)
        {
            // reset grid rows/cols and scroll bars if view changed
            if (viewChanged)
            {
                Clear(C1.Win.C1FlexGrid.ClearFlags.All);
                _calendarStart = DateTime.MinValue;

                if (_view == ViewEnum.Monthly)
                {
                    // set number of rows: current/nav, weekday names, 6 weeks
                    Rows.Count = 8;
                    Rows.Fixed = 2;

                    // set number of columns: 7 days
                    Cols.Count = 7;
                    Cols.Fixed = 0;

                    // no scroll bars, single cell selection, fixed-only merging
                    ScrollBars = ScrollBars.None;
                    SelectionMode = SelectionModeEnum.Cell;
                    AllowMerging = AllowMergingEnum.FixedOnly;
                    AllowEditing = false;
                }
                else if (_view == ViewEnum.Weekly)
                {
                    // set number of rows: current/nav, weekday names, 24 hours
                    Rows.Count = 26;
                    Rows.Fixed = 2;

                    // set number of columns: 7 days, time slots
                    Cols.Count = 8;
                    Cols.Fixed = 1;

                    // apply styles
                    ClearWeeklyView();

                    foreach (C1.Win.C1FlexGrid.Column c in Cols)
                    {
                        c.AllowMerging = true;
                        c.ComboList = "|...";
                    }

                    // vertical scroll bars, cell range selection, free merging
                    ScrollBars = ScrollBars.Vertical;
                    SelectionMode = SelectionModeEnum.CellRange;
                    AllowMerging = AllowMergingEnum.Free;
                    AllowEditing = true;
                }
            }

            // show/hide title, days of the week
            Rows[0].Visible = _btnPrev.Visible = _btnNext.Visible = _showTitle;
            Rows[1].Visible = _showDayNames;

            // get adjusted row count to calculate average row heights
            double cnt = (_view == ViewEnum.Monthly) ? Rows.Count - 2 : Rows.Count / 2;
            if (_showTitle) cnt++;
            if (_showDayNames) cnt++;

            // adjust average row/col sizes
            Rows[0].Height = Cols[0].Width = -1;
            Rows.DefaultSize = (int)(ClientSize.Height / cnt);
            Cols.DefaultSize = ClientSize.Width / Cols.Count;

            // fix round-off
            int delta = ClientSize.Height - Rows[Rows.Count - 1].Bottom;
            Rows[0].Height = Rows[0].HeightDisplay + delta;
            if (Rows[0].Height < 0) Rows[0].Height = 2 * Rows[0].HeightDisplay;
            delta = ClientSize.Width - Cols[Cols.Count - 1].Right;
            Cols[0].Width = Cols[0].WidthDisplay + delta;

            // set the first visible row to 6:00 am
            if (viewChanged && _view == ViewEnum.Weekly)
            {
                int top = 6;
                if (_showTitle) top++;
                if (_showDayNames) top++;
                TopRow = top;
            }

            // adjust button positions
            Rectangle rc = GetCellRect(0, 0, false);
            _btnPrev.Bounds = rc;
            _btnPrev.TextAlign = ContentAlignment.MiddleLeft;
            rc = GetCellRect(0, Cols.Count - 1, false);
            _btnNext.Bounds = rc;
            _btnNext.TextAlign = ContentAlignment.MiddleRight;
        }

        protected void BuildCalendar()
        {
            BuildCalendar(false);
        }

        protected void BuildCalendar(bool force)
        {
            // not while the mouse is down...
            if (!force && (Control.MouseButtons & MouseButtons.Left) != 0)
                return;

            // backtrack to first of the month
            DateTime date = SelectionStart;
            if (_view == ViewEnum.Monthly)
            {
                while (date.Day > 1)
                    date = date.AddDays(-1);

                date = date.AddDays(-1);
            }

            // backtrack to previous FirstDayOfWeek
            DayOfWeek first = GetFirstDayOfWeek();
            while (date.DayOfWeek != first)
                date = date.AddDays(-1);

            // same start date? no need to rebuild calendar
            if (!force && date == _calendarStart)
            {
                if (_view == ViewEnum.Weekly)
                {
                    SetWeeklyData(false); // user data needs to be reset
                }
                else
                {
                    Invalidate(); // repainting cells is sufficient
                    SyncSelection();
                }
                return;
            }

            _calendarStart = date;
            _calendarMonth = SelectionRange.Start;

            // column offset
            int offset = (_view == ViewEnum.Monthly) ? 0 : 1;

            // month on row 0
            Rows[0].Style = Styles["Title"];
            Rows[0].AllowMerging = true;
            CellRange rg = GetCellRange(0, 1, 0, 5 + offset);

            if (offset > 0)
            {
                rg.Data = "Week of " + _calendarStart.ToString("MMMM dd, yyyy");
            }
            else
            {
                rg.Data = _calendarMonth.ToString("MMMM, yyyy");
            }

            // DayNames on first row
            Rows[1].Style = (_view == ViewEnum.Monthly) ? Styles["DayNames"] : Styles["WeekDayNames"];
            DateTime weekDay = date;

            for (int c = 0; c < 7; c++)
            {
                this[1, c + offset] = weekDay;
                weekDay = weekDay.AddDays(1);
            }

            if (offset > 0)
            {
                this[1, 0] = null;
                weekDay = date;

                for (int i = 0; i < 24; i++)
                {
                    this[2 + i, 0] = weekDay;
                    weekDay = weekDay.AddHours(1);
                }

                CellRange cr = GetCellRange(1, 0, 25, 0);
                cr.Style = Styles["Hours"];

                cr = GetCellRange(0, 0);
                cr.Style = Styles["Title"];
                SetWeeklyData(true);
            }

            // invalidate all rows
            Invalidate();

            // synchronize grid cell with SelectionRange
            SyncSelection();
        }

        protected void SyncSelection()
        {
            DateTime t1 = GetCellDate(RowSel, ColSel);
            DateTime t2 = SelectionStart;
            int cmp = DateTime.Compare(t1, t2);

            if (cmp == 0)
                return;

            TimeSpan ts = t2 - t1;
            int n = System.Math.Abs(ts.Days);

            int newRow = RowSel;
            int newCol = ColSel;
            int maxCol = (cmp < 0) ? 6 : 0;
            int minCol = (cmp < 0) ? 0 : 6;

            while (n-- > 0)
            {
                if (newCol == maxCol)
                {
                    newRow -= cmp;
                    newCol = minCol;
                }
                else
                {
                    newCol -= cmp;
                }
            }

            if (_view == ViewEnum.Monthly)
            {
                Select(newRow, newCol);
            }
            else
            {
                Select(-1, -1);
            }
        }
        #endregion

        #region navigation buttons
        private void _btnPrevNext_Click(object sender, EventArgs e)
        {
            Focus();

            if (_view == ViewEnum.Monthly)
            {
                SelectionStart = SelectionStart.AddMonths(sender.Equals(_btnPrev) ? -1 : 1);

                if (DateChanged != null)
                    DateChanged(this, new System.EventArgs());
            }
            else
            {
                SelectionStart = SelectionStart.AddDays(sender.Equals(_btnPrev) ? -7 : 7);
            }

            BuildCalendar();
        }
        #endregion

        #region tooltips
        ToolTip _ttip;
        int _lastRow, _lastCol;

        private void _flex_MouseMoveTooltip(object sender, System.Windows.Forms.MouseEventArgs e)
        {
            string text = null;
            if (e.Button == MouseButtons.None)
            {
                // get mouse coordinates
                int row = this.MouseRow;
                int col = this.MouseCol;

                // save work if we can
                if (row == _lastRow && col == _lastCol)
                    return;

                // save info for next time
                _lastRow = row;
                _lastCol = col;

                // get text for tooltip
                if (row > 1 && col > 0)
                {
                    DateTime celldate = GetCellDate(row, col);
                    int lb = -1, ub = -1;

                    if (GetOwnerDrawTaskRange(celldate, ref lb, ref ub))
                    {
                        StringBuilder sb = new StringBuilder();

                        for (int i = lb; i <= ub; i++)
                        {
                            ITaskItem ti = _tasks[i] as ITaskItem;
                            sb.AppendFormat("\n{0:t} - {1}", ti.Start, ti.Task);
                        }

                        text = sb.ToString().Substring(1);

                        // get display rectangle
                        Rectangle rc = this.GetCellRect(row, col, false);
                        rc.Intersect(this.ClientRectangle);

                        // measure text
                        using (Graphics g = this.CreateGraphics())
                        {
                            CellStyle s = this.GetCellStyleDisplay(row, col);
                            float wid = g.MeasureString(text, s.Font).Width;
                            wid += s.Margins.Left + s.Margins.Right + s.Border.Width;
                        }
                    }
                }
            }

            // create tooltip if we need it
            if (text != null && _ttip == null)
            {
                _ttip = new ToolTip();
            }

            // set tooltip text
            if (_ttip != null && _ttip.GetToolTip(this) != text)
                _ttip.SetToolTip(this, text);
        }

        private void C1FlexCalendar_MouseMove(object sender, MouseEventArgs e)
        {
            if (_view == ViewEnum.Monthly)
                _flex_MouseMoveTooltip(sender, e);
        }
        #endregion
    }

    #region IComparer subclasses
    public class DailyTaskItemComparer : IComparer
    {
        int IComparer.Compare(object x, object y)
        {
            DateTime t1 = (x is ITaskItem) ? ((ITaskItem)x).Start : (DateTime)x;
            DateTime t2 = (y is ITaskItem) ? ((ITaskItem)y).Start : (DateTime)y;
            t1 = t1.Subtract(t1.TimeOfDay);
            t2 = t2.Subtract(t2.TimeOfDay);
            return DateTime.Compare(t1, t2);
        }
    }

    public class HourlyTaskItemComparer : IComparer
    {
        int IComparer.Compare(object x, object y)
        {
            if (x is ITaskItem && y is DateTime)
            {
                ITaskItem ti = (ITaskItem)x;
                DateTime dt = (DateTime)y;

                if (dt >= ti.Start && dt < ti.Finish)
                    return 0;
                else
                    return DateTime.Compare(ti.Start, dt);
            }
            else
            {
                DateTime t1 = (x is ITaskItem) ? ((ITaskItem)x).Start : (DateTime)x;
                DateTime t2 = (y is ITaskItem) ? ((ITaskItem)y).Start : (DateTime)y;
                return DateTime.Compare(t1, t2);
            }
        }
    }
    #endregion
}